###############################################################################
#
# Ferret search engine
#
# Written by Brian C. White <bcwhite@verisim.com>
# Copyright (c) 1996 by Verisim, Inc.  All rights reserved.
#
###############################################################################

package Ferret;

require 5.002;
require Exporter;
require DynaLoader;
require AutoLoader;

use Carp;

@ISA = qw(Exporter DynaLoader AutoLoader);
$VERSION   = '1.1';
@EXPORT	   = qw();
@EXPORT_OK = qw();

#$] == ===REQUIRED-PERL=== or die "Error: This instance of Ferret has been compiled against Perl v===REQUIRED-PERL===\n		  The currently installed version of Perl is v$]\n";

bootstrap Ferret $VERSION;



sub AUTOLOAD {
	my $constname = $AUTOLOAD;
	$constname =~ s/.*:://;
	my $val = constant($constname, @_ ? $_[0] : 0);
	if ($! != 0) {
#		if ($! =~ /Invalid/) {
			$AutoLoader::AUTOLOAD = $AUTOLOAD;
			goto &AutoLoader::AUTOLOAD;
#		} else {
#		  Carp::croak("Error: Nonexistent Ferret constant '$constname' used");
#		}
	}
	eval "sub $AUTOLOAD { $val }";
	goto &$AUTOLOAD;
}



use strict;



###############################################################################
#
# Package-wide variable definitions
#
my %indexmode;
my %newwords;
my %stoppers;
my %options;
my %picsizes;



###############################################################################
#
# Default list of stoppers (two letters or more only)
#
my @defstoppers =
	qw(and the with for who what where when why how her him them they
	   she its our ours that were will wont would wouldnt could cant
	   couldnt should shouldnt this that are from not can have all but
	   must mustnt these each also some has any use there which let);



###############################################################################
#
# Set of query aliases
#
my %queryaliases =
(
	AND		=> "&",
	OR		=> "|",
	NOT		=> "!",
	NEAR	=> "10",
);



###############################################################################
#
# Required string adjustments for documents & queries
#
sub AliasSubstitute {
	my $usage = 'Usage: Ferret::AliasSubstitute(\$string, \%aliases)';
	@_ == 2 or croak $usage;
	my($query,$aliases) = @_;

	croak $usage if (ref $query   ne "SCALAR");
	croak $usage if (ref $aliases ne "HASH");

#	print "alias in:  $$query\n";

	my($key,$val);
	while (($key,$val) = each %$aliases) {
		$$query =~ s/\b\Q$key\E\b/$val/gs;
	}

#	print "alias out: $$query\n";
}



sub PrepQuery {
	my $usage = 'Usage: Ferret::PrepQuery(\$query [, \$validwords])';
	@_ >= 1 && @_ <= 2 or croak $usage;
	my($query,$wlist) = @_;

	croak $usage if (ref $query ne "SCALAR");
	croak $usage if ($wlist && ref $wlist ne "SCALAR");

	# Remove any dangerous characters
	$$query =~ y/\$/ /s;
#	$$query =~ s/\$/ /g;

	# Remove any unknown characters
	$$query =~ y/A-Za-z0-9\' \t\n!|&()\"\[\]\?\*\^/ /c;
#	$$query =~ s/[^\w\s!|&()\"]+/ /gs;

	# Separate control characters out into words
	$$query =~ s/([!|&()])/ $1 /g;

	# Make sure quoted strings get evaluated properly
	$$query =~ s/"(.*?)"/ ( " $1 " ) /g;

	# Remove global word-matches
	1 while $$query =~ s/(^|\s)(\?|\*|\[\^.*?\])+($|\s)/ /gs;

	# Handle globbing patterns
	foreach (_FindGlobs($$query)) {
		my $expr = "( " . join(' | ',_ExpandGlob($_,$wlist)) . " )";
		$$query =~ s/\Q$_\E/$expr/;
	}

	# Merge all whitespace together
	$$query =~ s/\s+/ /gs;

	# Remove header and footer whitespace
	$$query =~ s/^\s+|\s+$//;

	# Mark entire query as subquery
	$$query = "( $$query )";
}



###############################################################################
#
# Opening & Closing Indexes
#
sub ReadOptions {
	my $usage = 'Usage: $ferret->ReadOptions()';
	@_ == 1 or croak $usage;
	my($this) = @_;

	croak $usage if (ref $this ne "Ferret");

	my $opts = $this->Options();
	$Ferret::options{$this}	 = $opts;

	if ($opts & &OPT_NOSTOPPERS) {
		$Ferret::stoppers{$this} = "";
	} else {
		$Ferret::stoppers{$this} = $this->DBGetWords("stoppers");
		$Ferret::stoppers{$this} = join("\n",@defstoppers) unless $Ferret::stoppers{$this};
	}
}



sub LocalOpen {
	my $usage = 'Usage: $ferret->LocalOpen(PATHNAME [, WRITEFLAG])';
	@_ >= 2 && @_ <= 3 or croak $usage;
	my($this,$file,$mode) = @_;

	croak $usage if (ref $this ne "Ferret");

	my $open = $this->OpenIndex($file,$mode);

	croak "Error: Could not open index file '$file' -- $!\n       stopped" if $open == -1;
	die	  "Error: Old index incompatible with newer evaluation version -- Please regenerate\n" if $open == -2;
	croak "Error: Evaluation version of \"Ferret\" has expired; stopped" if $open == -3;

	$Ferret::indexmode{$this}= $mode;
	$Ferret::newwords{$this} = "";
	$Ferret::stoppers{$this} = "";

	$this->ReadOptions();
}



sub Open {
	my $usage = 'Usage: $ferret->Open(PATHNAME)';
	@_ == 2 or croak $usage;
	my($this,$file) = @_;

	croak $usage if (ref $this ne "Ferret");

	$this->LocalOpen($file);
}



sub FlushLocal {
	my $usage = 'Usage: $ferret->FlushLocal()';
	@_ == 1 or croak $usage;
	my($this) = @_;

	croak $usage if (ref $this ne "Ferret");

	if ($Ferret::indexmode{$this} && $Ferret::newwords{$this}) {
		my $words = $this->DBGetWords("all");
		my @words = split(/\n/, $words);
		push @words,split(/\n/, $Ferret::newwords{$this});
		$words = join(' ',sort(@words));
		$this->DBPutWords("all",$words);
		$Ferret::newwords{$this} = "";
	}
}



sub Close {
	my $usage = 'Usage: $ferret->Close()';
	@_ == 1 or croak $usage;
	my($this) = @_;

	croak $usage if (ref $this ne "Ferret");

	$this->FlushLocal();
	$this->CloseIndex();
}



###############################################################################
#
# Queries
#
sub QueryRun {
	my $usage = 'Usage: $result = $ferret->QueryRun(QUERYSTRING)';
	@_ == 2 or croak $usage;
	my($this,$query) = @_;

	croak $usage if (ref $this ne "Ferret");

	my $wordlist = $this->DBGetWords("all") if $query =~ m/[\[\]\?\*]/;

	# Run standard aliases
	AliasSubstitute(\$query,\%queryaliases);

	StripWords(\$query, \$Ferret::stoppers{$this});
	PrepQuery(\$query,\$wordlist);

#	print "Prepped query=$query\n";

	while ($query =~ m/\(\s*([^()]*?)\s*\)/s) {
		my(@match,$lidx,$ridx,$index);
		my($op,$lr,$quote,$negate,$range);
		$negate = 0;

#		print "Subquery: $1\n";
#		_Time("Parsed Subquery");

		foreach (split(/ /,$1)) {
			if ($_ eq '"')		{ $quote = !$quote; next; }
			if ($_ eq '!')		{ $negate= 1;		next; }
			if ($_ =~ m/^\d+$/) { $range = $_;		next; }

			if ($_ =~ m/^[&|]$/) {
				$op = $_ if ($lr);
				next;
			}

			if ($_ =~ m/^\$([-\d]+)$/) {
				$ridx = $1;
			} else {
#				print "Query Word '$_'...";
				$ridx = $this->QueryWord($_);
#				print "(#$ridx)\n";
			}

			if ($negate) {
#				print "Negate Result #$ridx... ";
				$index = $this->QueryNegate($ridx);
#				print "(#$index)\n";
				$this->FreeQueryResult($ridx);
				$ridx = $index;
				$negate = 0;
			}
#			print "Match for '$_': @match\n";

			$range = 1 if $quote;

			if ($lr) {
				if ($op eq '|') {
					$index = $this->QueryOr($lidx,$ridx);
				} elsif ($op eq '' || $op eq '&') {
					$index = $this->QueryAnd($lidx,$ridx,$range);
				} else {
					$@ = "Unknown operator '$op'";
					return;
				}
				$this->FreeQueryResult($lidx);
				$this->FreeQueryResult($ridx);
				$lidx = $index;
			} else {
				$lidx = $ridx;
				$lr++;
			}

			$op = '';
		}

		$query =~ s/\(\s*[^()]*?\s*\)/\$$lidx/;
	}


	my($idx) = ($query =~ m/\$([-\d]+)/);
	return($idx);
}



sub QueryMatches {
	my $usage = 'Usage: @matches = $ferret->QueryMatches($result)';
	@_ == 2 or croak $usage;
	my($this,$idx) = @_;

	croak $usage if (ref $this ne "Ferret");

	my $docs = $this->QueryResults($idx);
	my @docs = split(/\n/,$docs);

	if (@docs) {
		@docs = sort { return $b <=> $a if ($b != $a); $a cmp $b } @docs;
		$@ = "";
	} else {
		$@ = "No matches found";
	}

	return(@docs);
}



sub Query {
	my $usage = 'Usage: @matches = $ferret->Query(QUERYSTRING)';
	@_ == 2 or croak $usage;
	my($this,$query) = @_;

	croak $usage if (ref $this ne "Ferret");

	my $result	= $this->QueryRun($query);
	my @matches = $this->QueryMatches($result);
	$this->FreeQueryResult($result);

	return @matches;
}



1;
__END__
###############################################################################
#
#
# Autoload methods go after __END__, and are processed by the autosplit program.
#
#
###############################################################################



###############################################################################
#
# General Utilities
#

sub Uniq {
	my $usage = 'Usage: @uniqlist = Ferret::Uniq(@sortedlist)';
	return () unless @_;

	croak $usage if (ref $_[0] eq "Ferret");

	my($last) = "";
	my(@new)  = ();
	foreach (@_) {
		next unless $_;
		if ($_ ne $last) {
			$last = $_;
			push(@new,$_);
		}
	}
	return @new;
}



sub RemoveFrom {
	my $usage = 'Usage: Ferret::RemoveFrom(\@fromlist, \@removelist)';
	@_ == 1 or croak $usage;
	my($flist,$rlist) = @_;

	croak $usage if (ref $flist ne "ARRAY");
	croak $usage if (ref $rlist ne "ARRAY");

	return unless @$flist && @$rlist;
	my $r=0; my $rmax=@$rlist; my $rval=$$rlist[0];
	my $f=0; my $fmax=@$flist; my $fval=$$flist[0];
	my @vals;
	LOOP: while ($f < $fmax) {
		while ($fval lt $rval) {
			push @vals,$fval;
			$f++;
			last LOOP unless ($f < $fmax);
			$fval = $$flist[$f];
		}
		if ($fval eq $rval) {
			$f++;
			last LOOP unless ($f < $fmax);
			$fval = $$flist[$f];
		}
		while ($fval gt $rval) {
			$r++;
			unless ($r < $rmax) {
				for (; $f < $fmax; $f++) {
					push @vals,$$flist[$f];
				}
				last LOOP;
			}
			$rval = $$rlist[$r];
		}
	}

	@$flist = @vals;
}



sub StdExtFilters {
	my $usage = 'Usage: $filter = Ferret::StdExtFilters(PATHNAME [, \$fixedpath])';
	@_ >= 1 && @_ <= 2 or croak $usage;
	my($file,$fixed) = @_;

	croak $usage if (ref $file eq "Ferret");
	croak $usage if ($fixed && ref $fixed ne "SCALAR");

	my($ffile);
	$fixed	= \$ffile unless $fixed;
	$$fixed =  $file;

	if ($file =~ s/\.gz$//)		{ $$fixed=$file; return "gzip -d";		}
	if ($file =~ s/\.Z$//)		{ $$fixed=$file; return "uncompress";	}

	return;
}



sub _Time {
	my $message = @_[0];
	my $curtime = time();

	if ($message) {
		my $diftime = $curtime - $Ferret::time;
		print "$message\: time=$diftime\n";
	}

	$Ferret::time = $curtime;
}



###############################################################################
#
# Time conversion utilities
#

sub TimetoRFC1123 {		#	Sun, 06 Nov 1994 08:49:37 GMT
	@_ == 1 or croak 'Usage: $date = Ferret::TimetoRFC1123(TIME)';
	my($time) = @_;

	croak $usage if (ref $time eq "Ferret");

	my @months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
	my @days   = qw(Sun Mon Tue Wed Thu Fri Sat);

	my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($time);
	$year += 1900 if $year < 1000;

	my $date = sprintf "%s, %02u %s %04u %02u:%02u:%02u GMT",$days[$wday],
			   $mday,$months[$mon],$year,$hour,$min,$sec;

	return $date;
}



###############################################################################
#
# General utility interfaces
#

sub DocumentList {
	@_ == 1 or croak 'Usage: @list = $ferret->DocumentList()';
	my($this) = @_;

	croak $usage if (ref $this ne "Ferret");

	my @list = split(/\n/,$this->ListDocuments());

	return @list;
}



###############################################################################
#
# More string adjustments for documents & queries
#
sub _GlobToRegex {
	my $glob  = @_[0];
	my $regex = $glob;
	$regex =~ s/\?/\\w/g;
	$regex =~ s/\*/\\w*/g;
#	print "$glob -> $regex\n";
	return($regex);
}



sub _FindGlobs {
	my $string = @_[0];
	my @globs  = grep(/[\?\*\[\]]/,split(/\s+/,$string));

	return (@globs);
}



sub _ExpandGlob {
	my $regex = _GlobToRegex(@_[0]);
	my $wlist = @_[1];
	my @words = ($$wlist =~ m/^($regex)$/gm);
	my $words = join("\n",@words);
	my @wlist;
	foreach $word (@words) {
		my $lcword = lc($word);
		next if ($lcword ne $word && $words =~ m/^$lcword$/m);
		push @wlist,$word;
	}

	return(@wlist);
}



###############################################################################
#
# URL utilities
#

sub ResolveURL {
	my $usage = 'Usage: $url = Ferret::ResolveURL(FROM-URL [, TO-URL])';
	@_ >= 1 && @_ <= 2 or croak $usage;
	my($from,$to) = @_;

	croak $usage if (ref $from eq "Ferret");

#	print "in: '$from' and '$to'\n";

	# Remove CGI arguments & tags
	$from =~ s/[?\#].*$//;

	my($ftype,$fhost,$fport,$ffile, $ttype,$thost,$tport,$tfile);
	($ftype,$fhost,$fport,$ffile) = ($from =~ m|^(\w+:)?(//[^/:]*)?(:\d+)?(.*)$|i);
	($ttype,$thost,$tport,$tfile) = ($to   =~ m|^(\w+:)?(//[^/:]*)?(:\d+)?(.*)$|i);

	$fhost =~ s!^//!!;
	$thost =~ s!^//!!;
	$ffile = "/"			unless $ffile;
	$tfile = $ffile			unless $to;
	$tfile = "/"			unless $tfile;
	$fport = ''				unless $fport;
	$tport = $fport			unless $tport && !$thost;
	$fhost = "localhost"	unless $fhost;
	$thost = $fhost			unless $thost;
	$ftype = "http:"		unless $ftype;
	$ttype = $ftype			unless $ttype;

#	print "type=$ftype, host=$fhost, port=$fport, file=$ffile -- ";
#	print "type=$ttype, host=$thost, port=$tport, file=$tfile\n";

	my $url,$file;
	if ($tfile =~ m|^/|) {
		$file = $tfile;
	} else {
		$file  = $ffile;
		$file =~ s|/[^/]+$|/|;
		$file .= $tfile;
	}
	$file =~ s|/+|/|g;
	1 while $file =~ s|/\./|/|g;
	1 while $file =~ s|/[^/]*/\.\./|/|g;

	$url  = "$ttype//$thost";
	$url .= "$tport" if ($tport);
	$url .= $file;

#	print "out: '$url'\n";
	return $url;
}



###############################################################################
#
# Document importing utilities
#

sub StripHTML {
	my $usage = 'Usage: Ferret::StripHTML(\$htmldata [, \$title [, \$summary [, SUMMARYSIZE]]])';
	@_ >= 1 && @_ <= 4 or croak $usage;
	my($doc,$title,$summary,$amount) = @_;

	croak $usage if (ref $doc ne "SCALAR");
	croak $usage if ($title && ref $title ne "SCALAR");
	croak $usage if ($summary && ref $summary ne "SCALAR");

	my($frameflag,$extrawords);
	$amount = 512 unless $amount;

	if ($title) {
		$$title = "";
		if ($$doc =~ m!<title>(.*?)</title>!is) {
			$$title = $1;
		}
	}

	$frameflag = ($$doc =~ m/<frame/);
	$extrawords= join(" ", ($$doc =~ m/<meta\s+(.*?)\s*>/gis));
	$extrawords=~s/\s+/ /gis;

	$$doc .= "\n" . $extrawords . "\n";

	$$doc =~ s!<head.*?>.*?</head.*?>!!gis;
	$$doc =~ s!<title.*?>.*?</title.*?>!!gis;
	$$doc =~ s|<!.*?>||gs;
	$$doc =~ s!<map.*?>.*?</map.*?>!!gis;
	$$doc =~ s!</?(html|body|meta|(no)?frame).*?>!!gis;
	$$doc =~ s!</?a(\s+.*?)?>!!gis;
	$$doc =~ s!^\s+!!;
	$$doc =~ s!\n+!\n!g;
	$$doc =~ s![ \t]+! !gs;

	if ($summary) {
#		$$doc =~ m/^(.{1,$amount})/s;
		$$summary =	 substr($$doc,0,$amount);
		$$summary =~ s/[^\n]*$//s;
		$$summary =~ s/<[^>]*$//;
		$$summary =~ s/\s*[^\s>]*$//;

		if ($frameflag) {
			$$summary = "[ This document uses frames -- the summary may be useless ] " . $$summary;
		}
	}

	$$doc .= $$title if $title;
	StripBadLines($doc);
	$$doc =~ s!<.*?>!!gs;
#	$$doc =~ s!\s+! !gs;
}



sub StripMIF {
	my $usage = 'Usage: Ferret::StripMIF(\$mifdata [, \$title [, \$summary [, SUMMARYSIZE]]])';
	@_ >= 1 && @_ <= 4 or croak $usage;
	my($doc,$title,$summary,$amount) = @_;

	croak $usage if (ref $doc ne "SCALAR");
	croak $usage if ($title && ref $title ne "SCALAR");
	croak $usage if ($summary && ref $summary ne "SCALAR");
	$amount = 512 unless $amount;

	if ($title) {
		$$title = "";
	}

	my @strings = ($$doc =~ m/<String\s+\`(.*?)\'>/gs);
	$$doc = join('\n',@strings);
	$$doc =~ s!<.*?>!!gs;

	if ($summary) {
#		$$doc =~ m/^(.{1,$amount})/;
		$$summary =	 substr($$doc,0,$amount);
		$$summary =~ s/[^\n]*$//s;
#		$$summary =~ s/\s+\S*$//s;
	}

	StripBadLines($doc);
	$$doc =~ s!\\x([0-9A-Fa-f]{2})!!gs;
	$$doc =~ s!\s+! !gs;
}



sub StripMSWord {
	my $usage = 'Usage: Ferret::StripMSWord(\$worddata [, \$title [, \$summary [, SUMMARYSIZE]]])';
	@_ >= 1 && @_ <= 4 or croak $usage;
	my($doc,$title,$summary,$amount) = @_;

	croak $usage if (ref $doc ne "SCALAR");
	croak $usage if ($title && ref $title ne "SCALAR");
	croak $usage if ($summary && ref $summary ne "SCALAR");
	$amount = 512 unless $amount;

#	s/^.*?([\w\W]

	my $ctrls = ord("S") - ord("@");
	my $ctrlt = ord("T") - ord("@");
	my $ctrlu = ord("U") - ord("@");
	my $sqopn = hex("91");
	my $sqcls = hex("92");
	my $dqopn = hex("93");
	my $dqcls = hex("94");

	s/$ctrls.{0,200}?$ctrlt(.{0,200}?)$ctrlu/$1/g;
	s/($sqopn|$sqcls)/'/g;	# '
	s/($dqopn|$dqcls)/"/g;	# "

	if ($title) {
		$$title = "";
	}

	if ($summary) {
		$$summary = "";
	}
}



sub StripText {
	my $usage = 'Usage: Ferret::StripText(\$textdata [, \$title [, \$summary [, SUMMARYSIZE]]])';
	@_ >= 1 && @_ <= 4 or croak $usage;
	my($doc,$title,$summary,$amount) = @_;

	croak $usage if (ref $doc ne "SCALAR");
	croak $usage if ($title && ref $title ne "SCALAR");
	croak $usage if ($summary && ref $summary ne "SCALAR");
	$amount = 512 unless $amount;

	if ($title) {
		$$title = "";
	}

	$$doc =~ s/^[\W\d\n]*\n//;
	$$doc =~ s/\n+/\n/g;

	if ($summary) {
#		$$doc =~ m/^(.{1,$amount})/s;
		$$summary =	 substr($$doc,0,$amount);
		$$summary =~ s/[^\n]*$//s;
	}

	StripBadLines($doc);
}



sub StripMail {
	my $usage = 'Usage: Ferret::StripMail(\$maildata [, \$title [, \$summary [, SUMMARYSIZE]]])';
	@_ >= 1 && @_ <= 4 or croak $usage;
	my($doc,$title,$summary,$amount) = @_;
	my($head);

	croak $usage if (ref $doc ne "SCALAR");
	croak $usage if ($title && ref $title ne "SCALAR");
	croak $usage if ($summary && ref $summary ne "SCALAR");
	$amount = 512 unless $amount;

	$head = substr($$doc,0,$amount+1000);
	if ($title) {
		$$title = "";
		($$title) = ($head =~ m/^subject:\s*(.*?)$/mi);
	}
#	_Time("Extract Title");

	# Remove MIME encoding information
	if ($$doc =~ m/^content-type:.*boundary="(.*?)"/mi) {
		my $mb = $1;
		$$doc =~ s/\n\s*-*$mb\s+.*?\n\s*\n/\n\n/gs;
	}
#	_Time("Remove MIME");

	if ($summary) {
		$$summary =  $head;
		$$summary =~ s/(^|\n)From .*?\n\s*\n/$1\n/sg;
#		$$summary =~ m/^(.{1,$amount})/s;
		$$summary =  substr($$summary,0,$amount);
		$$summary =~ s/[^\n]*$//s;
	}
#	_Time("Extract Summary");

#	$$doc =~ s/^[\W\d\n]*\n//;
#	$$doc =~ s/\n+/\n/g;

	StripBadLines($doc);
#	_Time("Remove Bad-Lines");
}



sub StripCode {
	my $usage = 'Usage: Ferret::StripCode(\$codedata [, \$title [, \$summary [, SUMMARYSIZE]]])';
	@_ >= 1 && @_ <= 4 or croak $usage;
	my($doc,$title,$summary,$amount) = @_;

	croak $usage if (ref $doc ne "SCALAR");
	croak $usage if ($title && ref $title ne "SCALAR");
	croak $usage if ($summary && ref $summary ne "SCALAR");
	$amount = 512 unless $amount;

	if ($title) {
		$$title = "";
	}

	$$doc =~ s/^#!.*\n//;
	$$doc =~ s/^[\W\d\n]*\n//;
	$$doc =~ s/\n+/\n/g;

	if ($summary) {
#		$$doc =~ m/^(.{1,$amount})/s;
		$$summary =	 substr($$doc,0,$amount);
		$$summary =~ s/[^\n]*$//s;
	}
}



sub ChangeHTMLForm {
	my $usage = 'Usage: Ferret::ChangeHTMLForm(\$htmldata, \$converthash)';
	@_ == 2 or croak $usage;
	my($doc,$conv) = @_;

	croak $usage if (ref $doc ne "SCALAR");
	croak $usage if (ref $conv ne "HASH");

	my $inputval = sub {
		my($type,$check) = @_;
		$type = lc($type);
		if ($check) {
			my($replace) = $type . "-checked";
			return $$conv{$replace} if $$conv{$replace};
		}
		return $$conv{$type};
	};

	$$doc =~ s!<(/?)form.*?>!$$conv{"$1form"}!gsi;
	$$doc =~ s!<select.*?>.*?($|<\s*/select.*?>)!$$conv{select}!gsi;
	$$doc =~ s!<textarea.*?>.*?($|<\s*/textarea.*?>)!$$conv{textarea}!gsi;
#	$$doc =~ s!<input.*?type="?(\w+)"?.*?>!print ">$1< "!gsie;
	$$doc =~ s!<input[^>]*type="?(\w+)"?[^>]*(checked).*?>!&$inputval($1,$2)!gsie;
	$$doc =~ s!<input[^>]*type="?(\w+)"?.*?>!&$inputval($1)!gsie;
}



sub MakeHTMLSummary {
	my $usage = 'Usage: Ferret::MakeHTMLSummary(\$summary, SOURCE-TYPE)';
	@_ == 2 or croak $usage;
	my($summary,$source) = @_;

	croak $usage if (ref $summary ne "SCALAR");

	if ($source =~ m/^html$/i) {
		$$summary =~ s!(<.*?)\scolor\s*=\s*#\w*(.*?>)!$1$2!gis;
		$$summary =~ s!<(/?h)(\d)>!"<$1" . ($2 > 5 ? 6 : $2 + 1) . ">"!gise;

		my $closers = "";
		foreach ($$summary =~ m/(<.*?>)/gs) {
			s/\s.*>$/>/s;

			/<(br|p|li|hr|img|dd|dt|input|meta|!.*)>/i and next;

			if (m|^</|) {
				$closers =~ s/^(.*?)\Q$_\E(.*)$/$1$2/i;
			} else {
				s|^<|</|;
				$closers = $_ . $closers;
			}
		}

		$$summary .= $closers;
	} else {
		QuoteHTML($summary);
		$$summary = "<pre>" . $$summary . "</pre>";
	}
}



sub LimitLineCount {
	my $usage = 'Usage: Ferret::LimitLineCount(\$data, LINECOUNT)';
	@_ == 2 or croak $usage;
	my($doc,$lines) = @_;

	croak $usage if (ref $doc ne "SCALAR");

	$$doc =~ s/^(([^\n]*\n){1,$lines})(.*)$/$1/s;
}



###############################################################################
#
# Image processing utilities
#
sub GetGIFSize {
	my $usage = 'Usage: Ferret::GetGIFSize(\$gif)';
	@_ == 1 or croak $usage;
	my($gif) = @_;

	croak $usage if (ref $gif ne "SCALAR");

	@_ = unpack ("C6v2", $$gif);
	return ($_[6], $_[7]);
}



sub GetJPGSize {
	my $usage = 'Usage: Ferret::GetJPGSize(\$jpeg)';
	@_ == 1 or croak $usage;
	my($jpeg) = @_;

	croak $usage if (ref $jpeg ne "SCALAR");

	my($c1,$c2,$len);
	my $jpos = 0;
	my $jlen = length $$jpeg;

	my $next_byte = sub {
		my $byte = substr($$jpeg,$jpos++,1);
		my $bval = unpack("C1",$byte);
		return $bval;
	};

	my $next_word = sub {
		my $b1 = &$next_byte();
		my $b2 = &$next_byte();
		return $b1 * 256 + $b2;
	};

	$c1 = &$next_byte();
	$c2 = &$next_byte();
	return if $c1 != 255 || $c2 != 216;	# first marker must be SOI

	while ($jpos < $jlen) {
		$c1 = &$next_byte();
		next if $c1 == 255;

		return if $c1==217 || $c1==218;	# stop at SOS or EOI

		if ($c1 >= 192 && $c1 <= 207) {
			if ($c1 != 196 && $c1 != 200 && $c1 != 204) {
				my($w,$h);
				   &$next_word();	# length
				   &$next_byte();	# precision
				$h=&$next_word();	# height
				$w=&$next_word();	# width
				return ($w,$h);
			}
		}

		$len = &$next_word();
		return if $len < 2;
		$len -= 2;
		while ($len--) { &$next_byte(); }
	}

	return;
}



sub ResolveHTMLImages {
	my $usage = 'Usage: Ferret::ResolveHTMLImages(\$html, FROM-URL [, SHRINK [, MAX-HEIGHT]])';
	@_ >= 2 && @_ <= 4 or croak $usage;
	my($doc,$url,$shrink,$maxheight) = @_;

	croak $usage if (ref $doc ne "SCALAR");

	$imgfix = sub {
		$_ = shift;
		my($source,$height,$width);

		s/\ssrc\s*=\s*\"?([^\s\">]+)\"?//i	and $source = $1;
		s/\sheight\s*=\s*([^\s>]+)//i		and $height = $1;
		s/\swidth\s*=\s*([^\s>]+)//i		and $width	= $1;

		$source = ResolveURL($url,$source);

		if ($shrink) {
			my $reduce = $shrink;
			unless ($height && $width) {
				if (exists $Ferret::picsizes{$source}) {
					($width,$height) = split(/,/,$Ferret::picsizes{$source});
#					print "Found cached '$source' ($width x $height)\n";
				} else {
					my($head,$data);
					unless ($@) {
						if ($source =~ m/\.gif$/i) {
							$data = eval { LoadURL($source,\$head) };
							($width,$height) = GetGIFSize(\$data);
						} elsif ($source =~ m/\.(jpg|jpeg)$/i) {
							$data = eval { LoadURL($source,\$head) };
							($width,$height) = GetJPGSize(\$data);
						}
					}
					if ($height && $width) {
						$Ferret::picsizes{$source} = "$width,$height";
#						print "Cached '$source' ($width x $height)\n";
					} else {
						$Ferret::picsizes{$source} = "$maxheight,$maxheight";
#						print "Cached '$source' (??? x ???)\n";
					}
				}
			}

			if ($maxheight && $height*$reduce > $maxheight) {
				$reduce	 = $maxheight / $height;
			}
			$height = int($height * $reduce);
			$width	= int($width  * $reduce);
		}

		if ($height && $width) {
			s/\s*>$/ src="$source" height=$height width=$width>/;
		} else {
			s/\s*>$/ src="$source">/;
		}

		return $_;
	};

	$$doc =~ s/(<img\s+[^>]*?>)/&$imgfix($1)/gise;
}



###############################################################################
#
# Document display utilities
#

sub GetSummary {
	@_ >= 2 or croak 'Usage: $ferret->GetSummary(DOCUMENTNAME [, ...])';
	my($this) = shift;

	croak $usage if (ref $this ne "Ferret");

	if (wantarray) {
		my @results = ();
		foreach (@_) {
			push @results, $this->DBGetSummary($_);
		}
		return @results;
	} else {
		return $this->DBGetSummary(shift @_);
	}
}



sub QueryWords {
	@_ == 1 or croak 'Usage: Ferret::QueryWords(QUERYSTRING)';
	my($query) = @_;

	croak $usage if (ref $query eq "Ferret");

	# Run standard aliases
	AliasSubstitute(\$query,\%queryaliases);

	my @words;
	$query =~ s/"\s*(.*?)\s*"/push @words,$1;""/gse;
	foreach (@words) {
		s/[^\w\[\]\?\*]+/ /gs;
		s/\s+/ /gs;
		y/A-Z/a-z/;
	}

	push @words, split(/[^\w\[\]\?\*]+/,$query);
	shift @words until $words[0];

	return @words;
}



sub QuoteHTML {
	my $usage = 'Usage: Ferret::QuoteHTML(\$htmldata)';
	@_ == 1 or croak $usage;
	my($doc) = @_;

	croak $usage if (ref $doc ne "SCALAR");

	$$doc =~ s|&|&amp;|g;
	$$doc =~ s|<|&lt;|g;
	$$doc =~ s|>|&gt;|g;
}



sub MarkTextWords {
	my $usage = 'Usage: Ferret::MarkTextWords(\$textdata, BEGINSTRING, ENDSTRING, WORD [, ...])';
	@_ >= 4 or croak $usage;
	my($doc,$begin,$end,@words) = @_;

	croak $usage if (ref $doc ne "SCALAR");

	foreach $word (@words) {
		my $patt = $word;
		my $case = $word eq lc($word);
		$patt =~ s/ /\\W+/gs;
		if ($case) {
			$$doc =~ s/\b($patt)\b/$begin$1$end/gsi;
		} else {
			$$doc =~ s/\b($patt)\b/$begin$1$end/gs;
		}
	}
}



sub MarkHTMLWords {
	my $usage = 'Usage: Ferret::MarkHTMLWords(\$document, BEGINSTRING, ENDSTRING, WORD [, ...])';
	@_ >= 4 or croak $usage;
	my($doc,$begin,$end,@words) = @_;
	my($g);

	croak $usage if (ref $doc ne "SCALAR");

	foreach $word (@words) {
		my $patt = $word;
		my $case = $word eq lc($word);

		foreach (_FindGlobs($patt)) {
			my $expr = _GlobToRegex($_);
			$patt =~ s/\Q$_\E/$expr/;
		}

		$patt =~ s/ /\\W+/gs;
		if ($case) {
			$$doc =~ s/(^|>)([^<]*)/$g=$1;$_=$2;s|\b($patt)\b|$begin$1$end|gsi;$g.$_/gse;
		} else {
			$$doc =~ s/(^|>)([^<]*)/$g=$1;$_=$2;s|\b($patt)\b|$begin$1$end|gs;$g.$_/gse;
		}
	}
}



###############################################################################
#
# More index management functions
#

sub Update {
	my $usage = 'Usage: $ferret->Update(PATHNAME)';
	@_ == 2 or croak $usage;
	my($this,$file) = @_;

	croak $usage if (ref $this ne "Ferret");

	$this->LocalOpen($file,1);
	Ferret::Message();
}



sub Flush {
	my $usage = 'Usage: $ferret->Flush()';
	@_ == 1 or croak $usage;
	my($this) = @_;

	croak $usage if (ref $this ne "Ferret");

	$this->FlushLocal();
	$this->FlushIndex();
}



sub Shrink {
	my $usage = 'Usage: $ferret->Shrink()';
	@_ == 1 or croak $usage;
	my($this) = @_;

	croak $usage if (ref $this ne "Ferret");

	$this->FlushLocal();
	$this->ShrinkIndex();
}



sub CopyTo {
	my $usage = 'Usage: $ferret->CopyTo(PATHNAME)';
	@_ == 2 or croak $usage;
	my($this,$file) = @_;

	croak $usage if (ref $this ne "Ferret");

	$this->FlushLocal();
	$this->CopyIndexTo($file);
}



###############################################################################
#
# Stopper Utilities
#

sub AddStoppers {
	my $usage = 'Usage: $ferret->AddStoppers(STOPPER [, STOPPER [, ...]])';
	@_ >= 2 or croak $usage;
	my($this,@stops) = @_;

	croak $usage if (ref $this ne "Ferret");

	return if ($Ferret::options{$this} & &OPT_NOSTOPPERS);

	my $stops = join(' ',@stops);
	StripNonWords(\$stops);
	StripWords(\$stops, \$Ferret::stoppers{$this});
	@stops = split(/ /,$stops);

	foreach (@stops) {
		$this->DBDel($_);
	}

	push @stops,split(/\n/,$Ferret::stoppers{$this});
	@stops = Uniq(sort(@stops));
	$Ferret::stoppers{$this} = join("\n",@stops);
	$this->DBPutWords("stoppers",$Ferret::stoppers{$this});
}



sub RemoveStoppers {
	my $usage = 'Usage: $ferret->RemoveStoppers(STOPPER [, STOPPER [, ...]])';
	@_ >= 2 or croak $usage;
	my($this,@remstops) = @_;

	croak $usage if (ref $this ne "Ferret");

	return if ($Ferret::options{$this} & &OPT_NOSTOPPERS);

	my @oldstops = split(/\n/,$Ferret::stoppers{$this});
	$Ferret::stoppers{$this} = "";

	my $stops = join(' ',@remstops);
	StripNonWords(\$stops);
	StripWords(\$stops, \$Ferret::stoppers{$this});
	@remstops = split(/ /,$stops);

	@remstops = Uniq(sort(@remstops));
	@oldstops = Uniq(sort(@oldstops));

#	print "Removing (@remstops) form (@oldstops)...\n";
	RemoveFrom(\@oldstops,\@remstops);

	$Ferret::stoppers{$this} = join("\n",@oldstops);
	$this->DBPutWords("stoppers",$Ferret::stoppers{$this});
}



###############################################################################
#
# Option management functions
#

sub SetOption {
	my $usage = 'Usage: $ferret->SetOption(Ferret::OPT_FLAG)';
	@_ == 2 or croak $usage;
	my($this,$flag) = @_;

	croak $usage if (ref $this ne "Ferret");

	my $result = $this->SetOptionFlag($flag);
	croak 'Error: Unknown option' if $result == -1;

	$this->ReadOptions();
}



sub UnsetOption {
	my $usage = 'Usage: $ferret->UnsetOption(Ferret::OPT_FLAG)';
	@_ == 2 or croak $usage;
	my($this,$flag) = @_;

	croak $usage if (ref $this ne "Ferret");

	my $result = $this->UnsetOptionFlag($flag);
	croak 'Error: Unknown option' if $result == -1;
	croak 'Error: Option cannot be unset; stopped' if $result == -2;

	$this->ReadOptions();
}



###############################################################################
#
# Index updates, deletes & adjustments
#

sub AddDocument {
	my $usage = 'Usage: $ferret->AddDocument(DOCNAME, DOCDATA [, BADWORDS [, ...]])';
	@_ >= 3 or croak $usage;
	my($this,$url,$doc,@strip) = @_;

	croak $usage if (ref $this ne "Ferret");

	my($record,$code,$counter,$olddoc);
	my(@list,@urlcodes);

	# Validate URL
	croak "Error: Document name '$url' cannot have embedded newlines; stopped" if ($url =~ m/\n/s);

#	_Time();
	# Remove any non-words (including digits)
	StripNonWords(\$doc);
#	_Time("Strip-Non-Words");

	# Remove any unsearchable words
	StripWords(\$doc, \$Ferret::stoppers{$this}, @strip);
#	_Time("Strip-Words");

	# Add the document to the database
	$Ferret::newwords{$this} .= $this->AddDocumentWords($url,$doc);
#	_Time("Add Words");
}



sub RemoveDocument {
	my $usage = 'Usage: $ferret->RemoveDocument(DOCNAME)';
	@_ == 2 or croak $usage;
	my($this,$url) = @_;

	croak $usage if (ref $this ne "Ferret");
	my $retcode = $this->RemoveDocumentWords($url);

	return ($retcode == 0);
}



sub MoveDocument {
	my $usage = 'Usage: $ferret->MoveDocument(OLDNAME, NEWNAME)';
	@_ == 3 or croak $usage;
	my($this,$old,$new) = @_;

	croak $usage if (ref $this ne "Ferret");

	my $result = $this->MoveDocumentWords($old,$new);

	croak "Error: Could not locate document '$old'; stopped" if $result < 0;
}



###############################################################################
#
# Document fetch utilities
#
sub FilterData {
	my $usage = 'Usage: Ferret::FilterData(\$data, FILTERPROG)';
	@_ == 2 or croak $usage;
	my($data,$filter) = @_;

	croak $usage if (ref $data ne "SCALAR");
	return unless $filter;

	my $tmpfile = "/tmp/url.$$";

	open(TMPFILE,">$tmpfile") || croak "Error: Could not create file '$tmpfile' -- $!; stopped";
	print TMPFILE $$data;
	close TMPFILE;

	my $delim = $/; undef $/;
	open(TMPFILE,"$filter <$tmpfile |");
	$$data = <TMPFILE>;
	close(TMPFILE);
	$/ = $delim;

	unlink($tmpfile);
}



sub LoadFile {
	my $usage = 'Usage: $data = Ferret::LoadFile(PATHNAME [, FILTERPROG])';
	@_ >= 1 && @_ <= 2 or croak $usage;
	my($file,$filter) = @_;

	croak $usage if (ref $file eq "Ferret");

#	_Time("Loading File '$file'...");
	croak "Error: Cannot read '$file'; stopped" unless -r $file;

	my $delim = $/; undef $/;
	if ($filter) {
		open(myLOADFILE,"$filter <$file |") || croak "Error: Could not filter '$file' with '$filter' -- $!; stopped";
	} else {
		open(myLOADFILE,"<$file") || croak "Error: Could not open '$file' -- $!; stopped";
	}
	my $data = <myLOADFILE>;
	close(myLOADFILE);
	$/ = $delim;

#	_Time("Loaded File '$file'");
	return($data);
}



sub LoadHTTP {
	my $usage = 'Usage: $data = Ferret::LoadHTTP(WEBFILE [, GET-OPTIONS])';
	@_ >= 1 && @_ <= 2 or croak $usage;
	my($addr,$opts) = @_;

	croak $usage if (ref $addr eq "Ferret");

	unless ($opts) {
		$opts  = "Connection: Keep-Alive\n";
	}

	my($http,$junk,$type,$host,$port);
	($http,$junk,$host,$port,$file) = ($addr =~ m|^(http:)?(//)?([^/:]*)?(:\d*)?(/.*)$|i);
	croak "Error: Badly formed HTTP address '$file'; stopped" unless $file;
	croak "Error: LoadHTTP cannot handle url in the form '$addr'; stopped" if ($http && $http !~ m/^http:$/i);
	$host = "localhost" unless $host;
	$port =~ s/^://;
	$port = 80 unless $port;

	use FileHandle;
	use Socket;

	my($proto,$sin,$ipaddr);

	$proto = getprotobyname('tcp');
	socket(SOCK, PF_INET, SOCK_STREAM, $proto);
	$ipaddr = inet_aton($host) || croak "Error: Could not locate machine '$host'; stopped";
	$sin = sockaddr_in($port,$ipaddr);
	connect(SOCK,$sin) || croak "Error: Could not connect to machine '$host' on port $port; stopped";

	my $delim = $/; undef $/;
	autoflush SOCK 1;
	print SOCK "GET $file HTTP/1.0\n$opts\n" || croak "Error: Could not write to socket -- $!; stopped";
	$data = <SOCK>;
	close(SOCK);
	$/ = $delim;

	croak "Error: No data from server for '$addr'; stopped" unless $data;

	return($data);
}



sub LoadURL {
	my $usage = 'Usage: $data = Ferret::LoadURL(URL, \$results [, FILTERFLAG])';
	@_ >= 2 && @_ <= 3 or croak $usage;
	my($url,$head,$dofilt) = @_;

	croak $usage if (ref $url eq "Ferret");
	croak $usage if (ref $head ne "SCALAR");

	my($file,$data,$filt);
	$@="";

	if ($url =~ m/^file:(.*)$/i) {
		$file = $1;
		$filt = StdExtFilters($file) if $dofilt;
		$$head= "";
		return LoadFile($file,$filt);
	} elsif ($url =~ m/^(http:)?(.*)$/i) {
		$file =	 $2;
		$data =	 LoadHTTP($file);
		$data =~ s/^(.*?\n)\s*\n//s;
		$$head=	 $1;
		$$head=~ s/\r//g;

		my($rnum,$rmsg) = ($$head =~ m/^HTTP\S*\s+(\d+)\s+(.*?)$/im);
		croak "Load $url failed: $rnum $rmsg; stopped" if ($rnum != 200);
	} else {
		croak "Error: Unsupported/unknown URL '$url'; stopped";
	}

	if ($dofilt) {
		$filt = StdExtFilters($file);
		FilterData(\$data,$filt) if ($filt);
	}

	return $data;
}



###############################################################################
#
# Remote Queries
#
sub RemoteQuery {
	my $usage = 'Usage: $data = Ferret::RemoteQuery(QUERY, REMOTE-URL-DIR [, ...] [, \@summaries])';
	@_ >= 2 or croak $usage;
	my $query = shift;
	my $sumref;

	croak $usage if (ref $query eq "Ferret");

	my $results = "";
	my $errors	= "";
	my $opts	= "";

	$query =~ s/\s+/+/g;
	$query = "remquery.cgi?1 " . $query;

	foreach $url (@_) {
		if (ref $url eq "ARRAY") {
			$sumref = $url;
			next;
		}

		my $data = eval { LoadHTTP($url.$query,$opts) };
		if ($@) {
			$errors .= $@ . "\n";
			next;
		}
		$data =~ s/\r//g;
		my($head) = ($data =~ m!^(.*?\n)\n!s);
#		my($type) = ($head =~ m!^Content-type:\s+(.*)!im);
		my($r,$m) = ($head =~ m!^HTTP\S*\s+(\d+)\s*(.*)$!im);
		next if ($r != 200);
		$data =~ s/^.*?\n\n//s;

		$results .= $data;
	}

#	print "Errors: '$errors', length(results)=".length($results)."\n";

	my(@matches,@summaries,@lines);

	@lines =split(/\n/,$results);
	while (@lines) {
		my $line = shift @lines;
		chop($line);
		next unless $line;
		push @matches,$line;

		$line = shift @lines;
		chop($line);
		my $size = $line;

		my $smry = "";
		while (@lines && length($smry) < $size) {
			$smry .= shift @lines;
		}
		push @summaries,$smry;
	}

	@$sumref = @summaries if (ref $sumref eq "ARRAY");
	return @matches;;
}
